"""
Entry point for easily exposing the useful imports
"""

from clothesline.real_interval_set import RealIntervalSet  # noqa: F401
from clothesline.enriched.datetime_interval_set import (  # noqa: F401, E501
    DatetimeIntervalSet,
)


def main():
    """Entry point for an application script"""
    print("clothesline offers no useful scripts.")
